# ==================================================================
# This R script illustrates the use of the package 'ergm' for the
# estimation and evaluation of exponential random graph models.
# ==================================================================
# 
# 
# The examples make use of classroom-based friendship nominations
# collected by Andrea Knecht in 2003/04 as part of her PhD work.
# 
# The R-packages 'ergm', 'network', 'coda' and 'sna' need to be
# available, . 
# 
# Overview: 
# a) Installation & reading data into R
# b) Some statistics and plots
# c) ERGM analysis of the classroom network
# d) ERGM analysis undirected 
# 
# 
# =====================================
# a) Installation & reading data into R
# =====================================
# 
# Install the package 'ergm' from the distributed ZIP archive:

install.packages("ergm")
install.packages("coda") # (might be part of 'ergm' already)
install.packages("latticeExtra")

# Load the vocabulary of the neccessary packages:

library(network) # network data storage
library(sna)     # network analysis routines
library(ergm)    # fitting & evaluating ERGMs
library(coda)    # MCMC diagnostics for goodness of fit
library(latticeExtra) # for nicer goodness of fit plots

# Set working directory to an appropriate location:

setwd("<...>")

# This needs to be adapted to a location where the files
# "klas03e-demographics.dat" and "klas03e-friends-waveC.dat"
# can be found.
# 
# Read the data, saved as ASCII files with adjacency matrix
# and the demographic characteristics sex, age, ethnicity & religion.

friendship <- as.matrix(read.table("klas03e-friends-waveC.dat"))
friendship[1:5,1:5]                                                        
dim(friendship)

demographics <- read.table("klas03e-demographics.dat",
  col.names=c("sexM","age","minority","religion")) 
demographics

# Transform the matrix into a network object and attach two of
# the actor variables to the network object:

friendnet <- network(friendship,directed=T)
friendnet %v% "sexM" <- demographics$sexM
friendnet %v% "religion" <- demographics$religion+3

# The use of the "+3" will become clear a bit further down...
# 
# =======================
# b) Statistics and plots
# =======================
# 
# The idea is to first get some impressions of the data by
# calculating descriptive statistics and making plots
# so we can run an "informed" ERGM.

plot(friendnet,vertex.col="sexM")

# The network is highly segregated according to sex.
# Let us recolour so it looks more clich-intuitive:

col.sexM <- rep("navy",length(demographics$sexM))
col.sexM[demographics$sexM==1] <- "tomato"

# This has created a vector to use in plot statement below

plot(friendnet,vertex.col=col.sexM)

# Now let's include religion as 'shape':

plot(friendnet,vertex.col=col.sexM,
  vertex.sides="religion",vertex.cex=2)

# The '+3' above was to render also missing religion info
# (code zero) visible as a shape: a triangle.
# 
# Histograms and plots of in and out degree, in one graphic:

par(mfrow=c(1,3))
odegree <- rowSums(friendship)
idegree <- colSums(friendship)
rdegree <- rowSums(symmetrize(friendship, rule="strong"))
hist(odegree)
hist(idegree)
hist(rdegree)

# The outdegree distribution could make one suspect outliers!
# Let us check... but first, to switch off the output screen, type

dev.off()

# Now plot the sociomatrix:

plot.sociomatrix(friendnet)

# This relieves fears of strong outliers.
# 
# What about the joint distribution of indegree, outdegree, and
# reciprocated degree?

par(mfrow=c(1,3))
plot(odegree,idegree)
plot(odegree,rdegree)
plot(idegree,rdegree)

# All seems positively related. However, a frequent finding for
# friendship at school is that there is hierarchy in the network
# as well. To show this, first switch off the graphics screen:

dev.off()

# Then calculate the correlation of indegree and outdegree: 

cor(odegree,idegree)

# A positive value - as one would expect, given the diagrams.
# It would suggest that 'popular' students (high indegree) are
# also 'active' students (high outdegree), and not a hierarchy
# between only-senders at the bottom and only-receivers on top.
# 
# However, this impression is contaminated by reciprocity!
# Let us calculate the correlation between the un-reciprocated
# parts of indegree and outdegree:

cor(odegree-rdegree,idegree-rdegree)

# Indeed a negative result! Senders of asymmetric friendship
# relations tend NOT to be receivers of such relations, and
# vice versa -- this indicates a (globally) hierarchical relation!

# Compare average degree & standard deviation between sexes:

tapply(odegree,demographics$sexM,mean)
tapply(odegree,demographics$sexM,sd)

# It seems the boys are more connected than the girls.
# 
# A bit more descriptives still about structural features:

gden(friendnet)
grecip(friendnet, measure="edgewise")
gtrans(friendnet)

# The following command gives a test of observed transitivity
# against the amount expected by simulations of conditionally
# random networks, given the correct (empirical) dyad census:

cug.test(friendnet,"gtrans",cmode="dyad.census")

# We can also take a look at this census, and the triad census:

dyad.census(friendnet)
triad.census(friendnet)

# Now let us look at two more 'global' features of the network.
# What does (1) the distribution of geodesic distances and (2) the
# shared partner distribution look like?
#
# Keep in mind these are "global" structural features we are trying 
# to also achieve with a good fitting ERG model - i.e., goodness of
# fit indicators!

distances <- geodist(friendnet)$gdist
distances[1:5,1:5]
hist(distances,xlab="directed distance between pairs")

# 'Edgewise shared partners' distribution (a measure of higher
# order clustering):

plot(0:10,summary(friendnet~esp(0:10)), type="b",
  xlab="number of edgewise shared partners", ylab="frequency")

# close graphics screen:

dev.off()

# =========================================
# c) ERGM analysis of the classroom network
# =========================================
# 
# The important help files are these:

?ergm
?'ergm-terms'
?mcmc.diagnostics
?gof

# Now finally, let us estimate some ERGMs for the classroom network!
# 
# Begin with a Bernoulli random graph model / Erds-Rnyi model:

model1 <- friendnet~edges
summary(model1)

# The summary method gives the "target statistics", here: the tie count.
# Let's estimate this model:

results1 <- ergm(model1)
summary(results1)

# The previous command takes a look at the model results.
# To see what else is there in the results, see

names(results1)

# Now as we do know that sex homophily matters from the plots, let us
# add the terms for "sexM" while still retaining dyadic independence.
# See ?'ergm-terms' for what in RSiena we'd call the "shortName"... 

model2 <- friendnet~edges+nodematch("sexM")
results2 <- ergm(model2)
summary(results2)
                                      
# See how the fit indices improve upon addition of sex homophily!
# 
# Now add reciprocity to the model & modify some estimation options:

model3 <- friendnet~edges+nodematch("sexM")+mutual
results3 <- ergm(model3,burnin=200000,MCMCsamplesize=10000)
summary(results3)

# Note how the sex homophily estimate changes.
# Everything highly significant... how do non-significant results
# look like? Include homophily on the religion variable...

model4 <- friendnet~edges+nodematch("sexM")+nodematch("religion")+mutual
results4 <- ergm(model4,burnin=200000,MCMCsamplesize=10000)
summary(results4)

# Based on the results, we could calculate conditional tie probabilites.
# Let us calculate the probabilty of a tie between a pair (i,j) who match
# on sex, mismatch on religion, and where the tie would reciprocate an
# already existing unilateral tie:
 
exp(sum(results4$coef*c(1,1,0,1)))/(1+exp(sum(results4$coef*c(1,1,0,1))))

# What the previous row does is to multiply estimates by the change
# statistics for the scenario sketched for pair (i,j), and then transform
# these log-odds into a probability.
#
# Same calculation, if it were not a reciprocating tie:

exp(sum(results4$coef*c(1,1,0,0)))/(1+exp(sum(results4$coef*c(1,1,0,0))))

# Much smaller. Once more the same calculation now assuming it is a
# reciprocating tie, but no same-sex tie:

exp(sum(results4$coef*c(1,0,0,1)))/(1+exp(sum(results4$coef*c(1,0,0,1))))

# All very nice, but did the model converge, actually?
# Let us look at the model diagnostics, i.e., the sample of network
# statistics that were simulated:

mcmc.diagnostics(results4,vars.per.page=4)

# This looks 'nice' (at least in my model run when writing this script) -
# there is no trend or cyclicity or weird pattern in the temporal plot of
# statistics; observed data are centrally located in simulated data.
# 
# Now let us look at the goodness of fit!
# Does the model reproduce some global features of the network?
# We check whether the empirical distribution of geodesic distances
# and shared partner counts can be explained by the model:

fit4 <- gof(results4~distance+espartners,nsim=40,burnin=200000)
par(mfrow=c(1,2)) # reserve two screens for two fit dimensions
plot(fit4) # plot those

# This is not a great fit (in my model run when writing this script),
# at least on shared partner counts. Also, there are less actor pairs
# at distance two than the model would predict.
# Both is not really a surprise, considering that no group formation
# (triangulation) tendencies were included in model 4. 
# Before going there, we give dyad-based modelling another try by
# allowing differential homophily tendencies of the two sexes:

model5 <- friendnet~edges+nodematch("sexM",diff=T)+
  nodematch("religion")+mutual
results5 <- ergm(model5,burnin=400000,MCMCsamplesize=30000)
summary(results5)

# Results (in my model run etc. etc.) don't look as if it mattered.
# AIC and BIC suggest I should not prefer model 5 over model 4.

mcmc.diagnostics(results5,vars.per.page=5)

# Convergence again is okay (in my...) - what about fit?

fit5 <- gof(results5~distance+espartners,nsim=40,burnin=200000)
par(mfrow=c(1,2))
plot(fit5)

# Not much changed, compared to previous model:

par(mfrow=c(2,2))
plot(fit4)
plot(fit5)

# As said before, bad fit is likely a result of not yet trying
# to capture the clustering in the network by a model coefficient.
# Let us introduce a clustering term: 'gwesp' (this is the alternating
# k-triangle statistic in 'ergm' language).

model6 <- friendnet~edges+nodematch("sexM")+mutual+gwesp(.5,fixed=T)
results6 <- ergm(model6,burnin=200000,MCMCsamplesize=10000)
summary(results6)

mcmc.diagnostics(results6,vars.per.page=4)

# If convergence is slightly off on any model dimension(s), try to
# reach better convergence with a new run using last estimates as
# new starting values:

results6a <- ergm(model6,theta0=results6$coef,burnin=200000,MCMCsamplesize=10000)

mcmc.diagnostics(results6a,vars.per.page=4)

# This can be iteratively continued until convergence is obtained.
# Sometimes, longer burnin phases may be required, or more steps
# between sampling from the Markov chain.
#
# What about (other) fit dimensions?

fit6 <- gof(results6~distance+espartners,nsim=40,burnin=200000)
fit6a <- gof(results6a~distance+espartners,nsim=40,burnin=200000)

par(mfrow=c(2,2))
plot(fit6)
plot(fit6a)

# In my case (when writing this script), run 6a gave a better convergence
# and also a better fit.

# If you want to see a run crash because of degeneracy issues, try using
# effect "transitive" instead of effect "gwesp" for modelling clustering.

# Final thing here: sender & receiver effects of sex (just for fun):

model7 <- friendnet~edges+nodematch("sexM")+nodeicov("sexM")+
          nodeocov("sexM")+mutual+gwesp(.5,fixed=T)
results7 <- ergm(model7,burnin=200000,MCMCsamplesize=10000)
summary(results7)
mcmc.diagnostics(results7,vars.per.page=6)

par(mfrow=c(2,2))
fit7 <- gof(results7~distance+espartners,nsim=40,burnin=200000)
plot(fit7)

# 
# ===========================
# d) ERGM analysis undirected 
# ===========================
# 
# We are now going to analyse Padgett's "Florentine families" data that
# are provided together with the ergm package. There data contains info
# about marriage and business relationships between families in early
# Renaissance Florence.

# the first part of the code comes from the ergm help files, b.t.w.

	 data(flo)

# attach the sociomatrix for the Florentine marriage data;
# let's make a network from the matrix:

     flomarriage <- network(flo,directed=FALSE)
     flomarriage

# create a vector indicating the wealth of each family (in thousands of lira) 
# and add it as a covariate to the network object

	 wealth <- c(10,36,27,146,55,44,20,8,42,103,48,49,10,48,32,3)
 
     flomarriage %v% "wealth" <- wealth

# create a plot of the social network

     plot(flomarriage)

# now make the vertex size proportional to their wealth

     plot(flomarriage, vertex.cex=wealth/10, main="Marriage Ties")

# Load a network object of the Florentine data

     data(florentine)

# Fit a model where the propensity to form ties between
# families depends on the absolute difference in wealth

     flo.model1 <- ergm(flomarriage ~ edges + absdiff("wealth"))
     summary(flo.model1)

	 x <- gof(flo.model1~distance,nsim=40,burnin=200000)
	 plot(x)

# Add triangle to model to see if there is clustering in the network

     flo.model2 <- ergm(flomarriage ~edges+absdiff("wealth")+
		triangle,burnin=100000)
     summary(flo.model2)

# diagnostics

	 mcmc.diagnostics(flo.model2,vars.per.page=3) 

# triangle seems a bit off (in my run...)
# what happens if we use gwesp instead of triangle

     flo.model3 <- ergm(flomarriage ~edges+absdiff("wealth")+
		gwesp(.5,fixed=T),burnin=100000)
     summary(flo.model3)

	 mcmc.diagnostics(flo.model3,vars.per.page=3)

# It appears gwesp hasn't converged (in my run...)

     flo.model4 <- ergm(flomarriage ~edges+absdiff("wealth")+
		gwesp(.5,fixed=T),burnin=500000,MCMCsamplesize=40000)
     summary(flo.model4)

	 mcmc.diagnostics(flo.model4,vars.per.page=3)

# probably not worth having a triangle or gwesp term (in my run...)

# now adding an edgecovariate-one can add an edge covariate in the form
# of a matrix or a network.
# here we put in a network of business relations between the families
# do families mix marriage with business?

     flo.model5 <- ergm(flomarriage ~edges+absdiff("wealth")+
		edgecov(flobusiness),burnin=100000)
     summary(flo.model5)  #adding edgecov

# goodness of Fit:

	 par(mfrow=c(1,2))
	 x <- gof(flo.model5~distance+espartners,nsim=40,burnin=200000)
	 plot(x)

# a very nice fit (in my run...) but do we need node covariates?

     flo.model6 <- ergm(flomarriage ~edges+absdiff("wealth")+ 
		edgecov(flobusiness)+nodecov("wealth"),burnin=100000)
     summary(flo.model6)

# how about k star terms?

     flo.model7 <- ergm(flomarriage ~edges+absdiff("wealth")+
		edgecov(flobusiness)+nodecov("wealth")+kstar(2:3),burnin=100000)
     summary(flo.model7)

	 mcmc.diagnostics(flo.model7,vars.per.page=5)

# looking at gof for distance, shared partners, and degree distribution 

	 par(mfrow=c(1,3))
	 x <- gof(flo.model7~distance+espartners+degree,nsim=40,burnin=200000)
	 plot(x)

# simulating one example network

	 sim.flo <- simulate(flo.model7,burnin=200000)

	 dev.off()
	 plot(sim.flo)

# maybe we do need a clustering term?

     flo.model8 <- ergm(flomarriage ~edges+absdiff("wealth") 
		+edgecov(flobusiness)+triangle,burnin=200000,MCMCsamplesize=20000)
     summary(flo.model8)  #adding triangle back into model

	 sim.flo <- simulate(flo.model7,burnin=200000)
	 dev.off()
	 plot(sim.flo)

	 par(mfrow=c(1,3))
	 x <- gof(flo.model8~distance+espartners+degree,nsim=100,burnin=200000)
	 plot(x) 

# doesn't seem to really improve fit in this somewhat open end...
